/*******************udpserver*****************/
/* Header files needed to use the sockets API. */
/* File contain Macro, Data Type and Structure */
/* definitions along with Function prototypes. */
/* header files */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "udpclientserv.h"

typedef struct server_info_s
{
    int sd;
} server_info_t;

typedef struct client_info_s
{
    client_id_t id;
    struct sockaddr_in  clientaddr;
} client_info_t;

#undef  ADDITEM
#define ADDITEM( _enum_cli_cmd_, _str_cli_cmd, _unused_comment_ )   [_enum_cli_cmd_]=_str_cli_cmd
const char * cli_command_strings[CLI_CMD_MAX] =
{
#include "cli_commands"
};
#undef ADDITEM

client_id_t generate_client_id(void * seed_data)
{
    struct sockaddr_in  * paddr = (struct sockaddr_in *) seed_data;
    client_id_t cid = 0;

/*
  uint32_t mask = ((UINT32_MAX<<24)>>24);
  cid = ntohl(paddr->sin_addr.s_addr)&mask;
  cid = cid * 100000;
  cid = cid + ntohs(paddr->sin_port);
*/
    cid = ntohs(paddr->sin_port);
    printf("UDP server - generated new client id: %u\n", cid);

    return cid;
}

int get_cmd_code(const char * cmd)
{
    int  cmd_idx = 0;

    for(cmd_idx = 0; cmd_idx < CLI_CMD_MAX; ++cmd_idx)
    {
        if(!strcmp(cmd,cli_command_strings[cmd_idx]))
        {
            break;
        }
    }

    return cmd_idx;
}

int bind_udp_socket(server_info_t * server, in_port_t port)
{
    assert(server);

    int rc;
    int sd;
    struct sockaddr_in  serveraddr;
    socklen_t           serveraddrlen = sizeof(serveraddr);

    /* The socket() function returns a socket */
    /* descriptor representing an endpoint. */
    /* The statement also identifies that the */
    /* INET (Internet Protocol) address family */
    /* with the UDP transport (SOCK_DGRAM) will */
    /* be used for this socket. */
    /******************************************/
    /* get a socket descriptor */
    if ((sd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
        perror("UDP server - socket() error");
        exit(-1);
    } else
        printf("UDP server - socket() is OK\n");

    printf("UDP server - try to bind...\n");

    /* After the socket descriptor is received, */
    /* a bind() is done to assign a unique name */
    /* to the socket. In this example, the user */
    /* set the s_addr to zero. This allows the */
    /* system to connect to any client that uses */
    /* port 1555. */
    /********************************************/
    /* bind to address */
    memset(&serveraddr, 0x00, serveraddrlen);
    serveraddr.sin_family = AF_INET;
    serveraddr.sin_port = htons(port);
    serveraddr.sin_addr.s_addr = htonl(INADDR_ANY);
    if ((rc = bind(sd, (struct sockaddr *)&serveraddr, serveraddrlen)) < 0) {
        perror("UDP server - bind() error");
        close(sd);
        /* If something wrong with socket(), just exit */
        exit(-1);
    } else
        printf("UDP server - bind() is OK\n");

    printf("UDP server - Using IP %s and port %d\n", inet_ntoa(serveraddr.sin_addr), SERVPORT);
    server->sd = sd;

    return 0;
}

int recv_udp_data(const server_info_t * server,
                  client_info_t * client,
                  char * buffer)
{
    assert(server);
    assert(client);
    assert(buffer);

    int rc;
    int                 buflen = MAX_TXT_MSG_SIZE;
    socklen_t           clientadrlen = sizeof(client->clientaddr);

    /* Use the recvfrom() function to receive the */
    /* data. The recvfrom() function waits */
    /* indefinitely for data to arrive. */
    /************************************************/
    /* This example does not use flags that control */
    /* the reception of the data. */
    /************************************************/
    /* Wait on client requests. */
    rc = recvfrom(server->sd, buffer, buflen, 0, (struct sockaddr *)&client->clientaddr, &clientadrlen);

    if (rc < 0) {
        perror("UDP Server - recvfrom() error");
        close(server->sd);
        exit(-1);
    } else
        printf("UDP Server - recvfrom() is OK...\n");

    printf("UDP Server - received from %s:%d the message:\n \"%s\"\n",
           inet_ntoa(client->clientaddr.sin_addr), ntohs(client->clientaddr.sin_port), buffer);

    return 0;
}

/*
 * decode request from client
 */
int decode_txt_msg(const char  *     msg,
                   client_id_t  *    p_out_cid,
                   msg_id_t      *   p_out_mid,
                   cli_cmd_enum_t *  p_out_cmd_code
    )
{
    assert(msg);
    assert(p_out_cid);
    assert(p_out_mid);
    assert(p_out_cmd_code);

    uint16_t   cid = 0;
    uint32_t   mid = 0;
    char       cmd_txt[MAX_CMD_SIZE+1] = { 0 };
    char       req_ch = 0;
    cli_cmd_enum_t cmd_code = 0;

    if(sscanf(msg, MSG_FMT_RD,
              &cid,
              &mid,
              cmd_txt,
              &req_ch
           ) == MSG_NUM_FLDS)
    {
        cmd_code = get_cmd_code(cmd_txt);
        if(CLI_CMD_MAX == cmd_code)
        {
            return -1;
        }

        if( '?' != req_ch )
        {
            return -1;
        }

        *p_out_cid = cid;
        *p_out_mid = mid;
        *p_out_cmd_code = cmd_code;
    }
    else
    {
        //TODO: fix this
        puts("... client request not understood ...");
        return -1;
    }

    return 0;
}

int process_client_req(client_info_t  *  client,
                       client_id_t       in_cid,
                       msg_id_t          in_mid,
                       cli_cmd_enum_t    in_cmd_code)
{
    assert(client);

    if(CLI_CMD_HELLO == in_cmd_code && !client->id && !in_cid && !in_mid )
    {
        in_cid = generate_client_id(&client->clientaddr);
        client->id = in_cid;
    }

    return 0;
}

/*
 * Create a response to send to cliet
 */
int encode_txt_msg(char       *      msg,
                   client_id_t       in_cid,
                   msg_id_t          in_mid,
                   cli_cmd_enum_t    in_cmd_code
    )
{
    assert(msg);

    memset(msg,'\0', MAX_TXT_MSG_SIZE);
    snprintf(msg,
             MAX_TXT_MSG_SIZE,
             MSG_FMT_WR,
             in_cid,
             in_mid,
             cli_command_strings[in_cmd_code],
             'Y'
        );

    return 0;
}

int send_udp_data(const server_info_t * server, client_info_t * client, char * buffer)
{
    assert(server);
    assert(client);
    assert(buffer);

    int rc;
    int                 buflen = strlen(buffer);
    socklen_t           clientadrlen = sizeof(client->clientaddr);

    /* Send a reply by using the sendto() function. */
    /* In this example, the system echoes the received */
    /* data back to the client. */
    /************************************************/
    /* This example does not use flags that control */
    /* the transmission of the data */
    /************************************************/
    /* Send a reply */
    printf("UDP Server - replying to the UDP client...\n");
    rc = sendto(server->sd, buffer, buflen, 0, (struct sockaddr *)&client->clientaddr, clientadrlen);
    printf("UDP Server - sent to %s:%d the message:\n \"%s\"\n",
           inet_ntoa(client->clientaddr.sin_addr), ntohs(client->clientaddr.sin_port), buffer);
    if (rc < 0) {
        perror("UDP server - sendto() error");
        close(server->sd);
        exit(-1);
    } else
        printf("UDP Server - sendto() is OK...\n");

    return 0;
}

int main(int argc, char *argv[])
{
    char buffer[MAX_TXT_MSG_SIZE] = { 0 };
    server_info_t server = { 0 };

    /* Run the server without argument */
    bind_udp_socket(&server, SERVPORT);

    client_info_t  client = { 0 };

    while(1)
    {
        client_id_t    cid;
        msg_id_t       mid;
        cli_cmd_enum_t cmd;
        printf("UDP server - Listening...\n");
        recv_udp_data(&server, &client, buffer);
        decode_txt_msg(buffer, &cid, &mid, &cmd);
        printf("UDP Server - decoded the message as:\n" MSG_FMT_WR "\n",
               cid, mid, cli_command_strings[cmd], '?');
        process_client_req(&client, cid, mid, cmd);
        encode_txt_msg(buffer, client.id, mid, cmd);
        send_udp_data(&server, &client, buffer);
    }

    /* When the data has been sent, close() the */
    /* socket descriptor. */
    /********************************************/
    /* //TODO:  close() the socket descriptor. on exit! */
    exit(0);
}
